// Trigram filter
// 
// Simplified code as example for a trigram filter.
//
// by feyd//godX.de

#include <stdint.h>
#include <vector>
#include <string>
#include <map>
#include <stdio.h>

// class holding trigrams and lists
class CTrigramFilter
{
public:
	CTrigramFilter() : m_nDocument(-1) {}
	virtual ~CTrigramFilter() {}

	int AddDocument(const char* pDocument)
	{
		m_nDocument++;
		
		size_t nLength = strlen(pDocument);
		if(nLength<3)
			return -1;
		
		nLength -= 2;
		for(size_t n=0; n<nLength; n++)
			AddTrigramDocument(GetTrigram(&pDocument[n]));

		return m_nDocument;
	}
	
	void FindDocument(const char* pSearch, std::vector<int>& output)
	{
		size_t nLength = strlen(pSearch);
		if(nLength<3)
			return;
		
		nLength -= 2;
		output.clear();
		bool bFirst = true;
		for(size_t n=0; n<nLength; n++)
		{
			std::map<int, std::vector<int> >::iterator it = m_trigrams.find(GetTrigram(&pSearch[n]));
			if(it==m_trigrams.end())
			{
				output.clear();
				return;
			}
			std::vector<int>& search = it->second;
			if(bFirst)
			{
				bFirst = false;
				output = search;
			} else {
				size_t nPosOutput = 0;
				size_t nPosSearch = 0;
				while(nPosOutput<output.size() && nPosSearch<search.size())
				{
					if(output[nPosOutput]<search[nPosSearch])
					{
						output.erase(output.begin()+nPosOutput);
					} else if(output[nPosOutput]>search[nPosSearch]) {
						nPosSearch++;
					} else {
						nPosSearch++;
						nPosOutput++;
					}
				}
				output.resize(nPosOutput);
			}
		}
	}
	
protected:
	int GetTrigram(const char* pTrigram)
	{
		return (((int)pTrigram[0]&0xff)<<16) | (((int)pTrigram[1]&0xff)<<8) | (((int)pTrigram[2]&0xff)<<0);
	}
	
	void AddTrigramDocument(int nTrigram)
	{
		std::vector<int>& documentList = m_trigrams[nTrigram];
		if(documentList.size()>0)
			if(documentList[documentList.size()-1]==m_nDocument)
				return;
		documentList.push_back(m_nDocument);
	} 
	int m_nDocument;
	std::map<int, std::vector<int> > m_trigrams;
};

// Application entry point (from libc)
int main(int argc, const char* argv[])
{
	printf("Trigram filter test...\r\n\r\n");

	if(argc>1)
	{
		CTrigramFilter trigramFilter;

		printf(" Input:\r\n");
		std::vector<const char*> strings(argc-2);
		for(int n=0; n<argc-2; n++)
		{
			strings[n] = argv[n+1];
			printf("  %d: %s\r\n", trigramFilter.AddDocument(strings[n]), strings[n]);
		}
		
		printf(" Filter:\r\n");
		std::vector<int> search;
		trigramFilter.FindDocument(argv[argc-1], search);
		if(search.size()>0)
		{
			for(size_t n=0; n<search.size(); n++)
				printf("  %d: %s\r\n", search[n], strings[search[n]]);
		} else {
			printf("  not found\r\n");
		}
	} else {
		printf("not enough arguments\r\n");
	}
	return 0;
}
